"""Forms for treebeard."""

from django import forms
from django.forms.models import modelform_factory as django_modelform_factory
from django.utils.html import conditional_escape
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

from treebeard.al_tree import AL_Node


class TreeNodeChoiceField(forms.ModelChoiceField):
    """
    ModelChoiceField for use with tree models. Applies indentation
    to the label to reflect the depth of the node.

    The queryset/choices provided to the field must be ordered
    in order for the choices to appear in the order that the tree is organised.
    """

    DEPTH_SEPARATOR = mark_safe("&nbsp;&nbsp;&nbsp;&nbsp;")

    def _get_indent(self, obj):
        return mark_safe(conditional_escape(self.DEPTH_SEPARATOR) * (obj.get_depth() - 1))

    def label_from_instance(self, obj):
        label = super().label_from_instance(obj)
        return mark_safe(conditional_escape(self._get_indent(obj)) + conditional_escape(label))


class MoveNodeForm(forms.ModelForm):
    """
    Form to handle moving a node in a tree.

    Handles sorted/unsorted trees.

    It adds two fields to the form:

    - Relative to: The target node where the current node will
                   be moved to.
    - Position: The position relative to the target node that
                will be used to move the node. These can be:

                - For sorted trees: ``Child of`` and ``Sibling of``
                - For unsorted trees: ``First child of``, ``Before`` and
                  ``After``

    .. warning::

        Subclassing :py:class:`MoveNodeForm` directly is
        discouraged, since special care is needed to handle
        excluded fields, and these change depending on the
        tree type.

        It is recommended that the :py:func:`movenodeform_factory`
        function is used instead.
    """

    __position_choices_sorted = (
        ("sorted-child", _("Child of")),
        ("sorted-sibling", _("Sibling of")),
    )

    __position_choices_unsorted = (
        ("first-child", _("First child of")),
        ("left", _("Before")),
        ("right", _("After")),
    )

    treebeard_position = forms.ChoiceField(required=True, label=_("Position"))
    treebeard_ref_node = TreeNodeChoiceField(
        required=False,
        label=_("Relative to"),
        empty_label=_("-- root --"),
        queryset=None,  # Populated in __init__
    )

    def _get_initial(self, instance):
        if self.is_sorted:
            position = "sorted-child"
            ref_node = instance.get_parent()
        else:
            prev_sibling = instance.get_prev_sibling()
            if prev_sibling:
                position = "right"
                ref_node = prev_sibling
            else:
                position = "first-child"
                if instance.is_root():
                    ref_node = None
                else:
                    ref_node = instance.get_parent()
        return {"treebeard_ref_node": ref_node, "treebeard_position": position}

    def _set_ref_model_queryset(self, opts, instance):
        """
        Sets the queryset (or choices) on the treebeard_ref_model field.

        For AL trees, this sets a list of nodes as the choices. For all other trees,
        sets a queryset.

        Excludes the instance and its descendants since a move relative to those would be invalid
        """
        if issubclass(opts.model, AL_Node):
            choices = opts.model.get_tree()
            descendants = instance.get_descendants(include_self=True) if instance else []
            field = self.fields["treebeard_ref_node"]
            self.fields["treebeard_ref_node"]._choices = [("", "--------")] + [
                (field.prepare_value(node), field.label_from_instance(node))
                for node in choices
                if node not in descendants
            ]
            # Must set queryset so that the manually defined choices are valid.
            # Must also do this *after* setting _choices above, otherwise the setter
            # for queryset will overwrite the choices.
            self.fields["treebeard_ref_node"].queryset = opts.model.objects.all()
            return

        queryset = opts.model.get_tree()
        descendants = instance.get_descendants(include_self=True) if instance else None
        if descendants:
            queryset = queryset.exclude(pk__in=descendants.values_list("pk", flat=True))

        self.fields["treebeard_ref_node"].queryset = queryset

    def __init__(self, *args, initial=None, instance=None, **kwargs):
        opts = self._meta
        if opts.model is None:
            raise ValueError("ModelForm has no model class specified")

        self.is_sorted = getattr(opts.model, "node_order_by", False)

        # Set initial values for treebeard fields

        initial_ = self._get_initial(instance) if instance else {}
        if initial is not None:
            initial_.update(initial)

        super().__init__(*args, instance=instance, initial=initial_, **kwargs)

        # update the 'treebeard_position' field choices
        self.fields["treebeard_position"].choices = (
            self.__position_choices_sorted if self.is_sorted else self.__position_choices_unsorted
        )

        self._set_ref_model_queryset(opts, instance)

    def save(self, commit=True):
        """
        Saves the model form.

        WARNING: Treebeard does not respect commit=False: other nodes that
        need to be modified to make space for the edited node will be updated
        in the database, and thus it is difficult to avoid writing any changes to the database.

        TreeAdmin handles this by rolling back the entire transaction if the form or any
        inlines report an error. If you use this form elsewhere, you will need to do the same.
        """

        reference_node = self.cleaned_data.pop("treebeard_ref_node", None)
        position_type = self.cleaned_data.pop("treebeard_position")

        if self.instance._state.adding:
            if reference_node:
                self.instance = reference_node.add_child(instance=self.instance)
                self.instance.move(reference_node, pos=position_type)
            else:
                self.instance = self._meta.model.add_root(instance=self.instance)
        else:
            self.instance.save()
            if reference_node:
                self.instance.move(reference_node, pos=position_type)
            else:
                pos = "sorted-sibling" if self.is_sorted else "first-sibling"
                self.instance.move(self._meta.model.get_first_root_node(), pos)
        # Reload the instance
        self.instance.refresh_from_db()
        super().save(commit=commit)
        return self.instance


def movenodeform_factory(model, form=MoveNodeForm, exclude=None, **kwargs):
    """Dynamically build a MoveNodeForm subclass with the proper Meta.

    :param Node model:

        The subclass of :py:class:`Node` that will be handled
        by the form.

    :param form:

        The form class that will be used as a base. By
        default, :py:class:`MoveNodeForm` will be used.

    Accepts all other kwargs that can be passed to Django's `modelform_factory`.

    :return: A :py:class:`MoveNodeForm` subclass
    """
    if exclude is None:
        exclude = ()
    exclude += getattr(model, "MOVENODE_FORM_EXCLUDED_FIELDS", ())
    return django_modelform_factory(model, form, exclude=exclude, **kwargs)



"""Postgres Ltree Trees"""

import functools
import itertools
import operator
import string
from collections.abc import Iterable
from typing import Any

from django.core import serializers
from django.db import models, transaction
from django.db.models import F, Func, OuterRef, Q, Subquery, Value
from django.db.models.functions import Concat
from django.utils.translation import gettext_noop as _

from treebeard.exceptions import InvalidMoveToDescendant, NodeAlreadySaved, PathOverflow
from treebeard.models import Node

from .fields import Ltree2Text, PathField, PathValue, Subpath, Text2LTree


class InvalidLabelConstraints(Exception): ...


def generate_label(
    skip: set[str],
    before: str | None = None,
    after: str | None = None,
):
    """
    Generate a new label value that will order the label before `before` and after `after`.

    Uses an alphabet of digits and ascii uppercase letters.
    If no `before` constraint is provided, then chooses only from letters: this allows room
    for digits to be used in future if nodes are inserted to the left of this one, without
    having to move large chunks of the tree.

    :raise InvalidLabelConstraints: when no label could be generated within the given constraints.
    """
    char_choices = string.ascii_uppercase if not before and not after else string.digits + string.ascii_uppercase

    start = after or char_choices[0]

    if before and before <= start:
        raise InvalidLabelConstraints

    # Construct sets of characters for each position, appending one at the end if we need to extend the string
    char_lists = [char_choices[idx:] for idx in [char_choices.index(char) for char in start]]
    char_lists.append(char_choices)

    # There is no point testing portions of the strings that are identical
    start_from = 0
    for ch1, ch2 in zip(before or "", after or ""):
        if ch1 != ch2:
            break
        start_from += 1

    for i in range(start_from, len(char_lists)):
        iter = itertools.product(*char_lists[: i + 1])
        for label_parts in iter:
            label = "".join(label_parts)
            if after and label <= after:
                continue
            if label in skip:
                continue
            if before and label >= before:
                break  # No point looking any further
            return label

    # We should never reach here... right?
    raise ValueError("Failed to generate label. Please report this as a bug.")


def generate_path(
    prefix: PathValue | None = None,
    before: str | None = None,
    after: str | None = None,
    skip: Iterable[PathValue] | None = None,
) -> PathValue:
    skip = {path[-1] for path in skip} if skip else set()
    label = generate_label(before=before, after=after, skip=skip)
    return PathValue((prefix or []) + [label])


class LT_NodeQuerySet(models.query.QuerySet):
    """
    Custom queryset for the tree node manager.

    Needed only for the custom delete method.
    """

    def delete(self, *args, **kwargs):
        """
        Custom delete method, will remove all descendant nodes to ensure a
        consistent tree (no orphans)

        :returns: tuple of the number of objects deleted and a dictionary
                  with the number of deletions per object type
        """
        # Construct the minimal list of nodes that need deleting along with their descendants
        paths_to_remove = set()
        for node in self.order_by("path").only("path").iterator():
            found = False
            for depth in range(1, len(node.path)):
                path = node.path[0:depth]
                if str(path) in paths_to_remove:
                    # we are already removing an ancestor of this node, so skip it
                    found = True
                    break
            if not found:
                paths_to_remove.add(str(node.path))

        model = self.model.tree_model()

        if not paths_to_remove:
            return super(LT_NodeQuerySet, model.objects.none()).delete(*args, **kwargs)

        query = functools.reduce(operator.or_, [Q(path__descendants=path) for path in paths_to_remove])
        return super(LT_NodeQuerySet, model.objects.filter(query)).delete(*args, **kwargs)

    delete.alters_data = True
    delete.queryset_only = True


class LT_NodeManager(models.Manager):
    """Custom manager for nodes in a Materialized Path tree."""

    def get_queryset(self):
        """Sets the custom queryset as the default."""
        return LT_NodeQuerySet(self.model).order_by("path")


class LT_ComplexAddMoveHandler:
    def _get_new_path(self, prefix, insert_before, insert_after, nodes_moved=False, num_iters=0):
        insert_before_path = insert_before.path[-1] if insert_before else None
        insert_after_path = insert_after.path[-1] if insert_after else None

        if insert_before_path and insert_after_path and insert_after_path > insert_before_path:
            raise ValueError("Invalid path constraints")

        siblings = self.node_cls.tree_model().objects.filter(path__descendants=prefix, path__depth=len(prefix) + 1)

        try:
            path = generate_path(
                prefix=prefix,
                before=insert_before_path,
                after=insert_after_path,
                skip=siblings.values_list("path", flat=True),
            )
        except InvalidLabelConstraints:
            # We failed to find a path within the available range. That means we need to move the insert_before
            # node and everything after it to the right
            self._move_subtree_right(insert_before)
            insert_before.refresh_from_db()
            num_iters += 1

            # We should never end up here... the move right operation should always succeed
            if num_iters > 2:
                raise PathOverflow("Failed to generate a new path for node. Please report this as a bug.")
            return self._get_new_path(prefix, insert_before, insert_after, nodes_moved=True, num_iters=num_iters)

        return path, nodes_moved

    def _move_subtree_right(self, start_node):
        """
        Move the node and everything after it in the tree to the right. This is achieved simply by
        appending an extra character (A) to the topmost label in the path.
        """
        result_class = self.node_cls.tree_model()
        node_depth = len(start_node.path)
        if node_depth > 1:
            result_class.objects.filter(path__gte=start_node.path, path__depth=node_depth).update(
                path=Concat(
                    Subpath(F("path"), 0, node_depth - 1),
                    Text2LTree(Concat(Ltree2Text(Subpath(F("path"), node_depth - 1, 1)), Value("A"))),
                )
            )
            result_class.objects.filter(path__gte=start_node.path, path__depth__gt=node_depth).update(
                path=Concat(
                    Subpath(F("path"), 0, node_depth - 1),
                    Text2LTree(Concat(Ltree2Text(Subpath(F("path"), node_depth - 1, 1)), Value("A"))),
                    Subpath(F("path"), node_depth),
                ),
            )
        else:
            # Moving root nodes
            result_class.objects.filter(path__gte=start_node.path, path__depth=1).update(
                path=Text2LTree(Concat(Ltree2Text(Subpath(F("path"), 0, 1)), Value("A")))
            )
            result_class.objects.filter(path__gte=start_node.path, path__depth__gt=1).update(
                path=Concat(Text2LTree(Concat(Ltree2Text(Subpath(F("path"), 0, 1)), Value("A"))), Subpath(F("path"), 1))
            )


class LT_AddRootHandler:
    def __init__(self, cls, **kwargs):
        super().__init__()
        self.cls = cls
        self.kwargs = kwargs

    def process(self):
        # Lock all root node rows to avoid integrity errors. We must force evaluation of the queryset
        list(self.cls.get_root_nodes().select_for_update().only("pk"))

        # do we have a root node already?
        last_root = self.cls.get_last_root_node()

        if last_root and last_root.node_order_by:
            # There are root nodes and node_order_by has been set.
            # Delegate sorted insertion to add_sibling.
            # We must pass an instance here to ensure that the right object is created for
            # models with multi-table inheritance.
            return last_root.add_sibling(
                "sorted-sibling", instance=self.kwargs.get("instance") or self.cls(**self.kwargs)
            )

        if len(self.kwargs) == 1 and "instance" in self.kwargs:
            # adding the passed (unsaved) instance to the tree
            newobj = self.kwargs["instance"]
            if not newobj._state.adding:
                raise NodeAlreadySaved("Attempted to add a tree node that is already in the database")
        else:
            # creating the new object
            newobj = self.cls(**self.kwargs)

        if last_root:
            # adding the new root node as the last one
            newobj.path = generate_path(skip=self.cls.get_root_nodes().values_list("path", flat=True))
        else:
            # adding the first root node
            newobj.path = generate_path()

        # saving the instance before returning it
        newobj.save()
        return newobj


class LT_AddChildHandler:
    def __init__(self, node, creation_kwargs: dict[str, Any]):
        super().__init__()
        self.node = node
        self.node_cls = node.__class__
        # These are deliberately not extracted in the function signature to avoid collision with model field names
        self.kwargs = creation_kwargs

    def process(self):
        # Lock parent row
        self.node_cls.tree_model().objects.filter(pk=self.node.pk).select_for_update().only("pk").get()
        if self.node_cls.node_order_by and not self.node.is_leaf():
            # there are child nodes and node_order_by has been set
            # delegate sorted insertion to add_sibling
            return self.node.get_last_child().add_sibling("sorted-sibling", **self.kwargs)

        if len(self.kwargs) == 1 and "instance" in self.kwargs:
            # adding the passed (unsaved) instance to the tree
            newobj = self.kwargs["instance"]
            if not newobj._state.adding:
                raise NodeAlreadySaved("Attempted to add a tree node that is already in the database")
        else:
            # creating a new object
            newobj = self.node_cls(**self.kwargs)

        if self.node.is_leaf():
            # the node had no children, adding the first child
            newobj.path = generate_path(prefix=self.node.path)
        else:
            # adding the new child as the last one
            newobj.path = generate_path(
                prefix=self.node.path,
                skip=self.node.get_children().values_list("path", flat=True),
                after=self.node.get_children().last().path[-1],
            )

        # saving the instance before returning it
        newobj.save()
        return newobj


class LT_AddSiblingHandler(LT_ComplexAddMoveHandler):
    def __init__(self, node, pos, creation_kwargs: dict[str, Any]):
        super().__init__()
        self.node = node
        self.node_cls = node.__class__
        self.pos = pos
        # These are deliberately not extracted in the function signature to avoid collision with model field names
        self.kwargs = creation_kwargs

    def process(self):
        self.pos = self.node._prepare_pos_var_for_add_sibling(self.pos)

        if len(self.kwargs) == 1 and "instance" in self.kwargs:
            # adding the passed (unsaved) instance to the tree
            newobj = self.kwargs["instance"]
            if not newobj._state.adding:
                raise NodeAlreadySaved("Attempted to add a tree node that is already in the database")
        else:
            # creating a new object
            newobj = self.node_cls(**self.kwargs)

        insert_before = None
        insert_after = None
        if self.pos == "sorted-sibling":
            insert_before = self.node.get_sorted_pos_queryset(self.node.get_siblings(), newobj).first()
            if insert_before:
                insert_after = insert_before.get_prev_sibling()
        elif self.pos == "first-sibling":
            insert_before = self.node.get_siblings().first()
        elif self.pos == "left":
            insert_before = self.node
            insert_after = self.node.get_prev_sibling()
        elif self.pos == "right":
            insert_after = self.node
            insert_before = self.node.get_next_sibling()
        elif self.pos == "last-sibling":
            insert_after = self.node.get_siblings().last()

        newobj.path, _ = self._get_new_path(self.node.path[:-1], insert_before, insert_after)
        newobj.save()
        return newobj


class LT_MoveHandler(LT_ComplexAddMoveHandler):
    def __init__(self, node, target, pos=None):
        super().__init__()
        self.node = node
        self.node_cls = node.__class__
        self.target = target
        self.pos = pos

    def process(self):
        self.pos = self.node._prepare_pos_var_for_move(self.pos)

        if self.pos in ("first-child", "last-child", "sorted-child"):
            if self.target == self.node:
                raise InvalidMoveToDescendant(_("Can't move node to itself."))

        if self.target.is_descendant_of(self.node):
            raise InvalidMoveToDescendant(_("Can't move node to a descendant."))

        insert_before = None
        insert_after = None
        prefix = self.target.path[:-1]  # Assume initially that target is a sibling

        if self.pos == "sorted-sibling":
            insert_before = self.node.get_sorted_pos_queryset(self.target.get_siblings(), self.node).first()
            if insert_before:
                insert_after = insert_before.get_prev_sibling()
        elif self.pos == "sorted-child":
            insert_before = self.node.get_sorted_pos_queryset(self.target.get_children(), self.node).first()
            if insert_before:
                insert_after = insert_before.get_prev_sibling()
            prefix = self.target.path
        elif self.pos == "first-sibling":
            insert_before = self.target.get_siblings().first()
        elif self.pos == "last-sibling":
            insert_after = self.target.get_siblings().last()
        elif self.pos == "left":
            insert_before = self.target
            insert_after = self.target.get_prev_sibling()
        elif self.pos == "right":
            insert_after = self.target
            insert_before = self.target.get_next_sibling()
        elif self.pos == "first-child":
            insert_before = self.target.get_children().first()
            prefix = self.target.path
        elif self.pos == "last-child":
            insert_after = self.target.get_children().last()
            prefix = self.target.path

        new_path, nodes_moved = self._get_new_path(prefix, insert_before, insert_after)
        if nodes_moved:
            self.node.refresh_from_db()

        # Update the path for all the descendants of the node
        result_class = self.node_cls.tree_model()
        result_class.objects.filter(path__descendants=self.node.path, path__depth__gt=len(self.node.path)).update(
            path=Concat(
                Value(new_path, output_field=PathField()),
                Subpath(F("path"), len(self.node.path)),
            )
        )
        # And update the path for the node itself
        self.node.path = new_path
        self.node.save()


class LT_Node(Node):
    """Abstract model to create your own Postgres LTree trees."""

    node_order_by = []
    path = PathField(unique=True)

    TREEBEARD_IDENTIFYING_FIELD = "path"
    MOVENODE_FORM_EXCLUDED_FIELDS = ("path",)

    objects = LT_NodeManager()

    _cached_attributes = (*Node._cached_attributes,)

    @classmethod
    @transaction.atomic
    def add_root(cls, **kwargs):
        """
        Adds a root node to the tree.

        This method saves the node in database. The object is populated as if via:

        ```
        obj = cls(**kwargs)
        ```
        """
        return LT_AddRootHandler(cls, **kwargs).process()

    @classmethod
    def dump_bulk(cls, parent=None, keep_ids=True):
        """Dumps a tree branch to a python data structure."""

        cls = cls.tree_model()

        # Because of fix_tree, this method assumes that the depth
        # and numchild properties in the nodes can be incorrect,
        # so no helper methods are used
        qset = cls.objects.all()
        if parent:
            qset = qset.filter(path__descendants=parent.path)
        ret, lnk = [], {}
        pk_field = cls._meta.pk.attname
        for pyobj in serializers.serialize("python", qset.iterator()):
            # django's serializer stores the attributes in 'fields'
            fields = pyobj["fields"]
            path = PathValue(fields["path"])
            depth = len(path)
            # this will be useless in load_bulk
            del fields["path"]
            if pk_field in fields:
                # this happens immediately after a load_bulk
                del fields[pk_field]

            newobj = {"data": fields}
            if keep_ids:
                newobj[pk_field] = pyobj["pk"]

            if (not parent and depth == 1) or (parent and len(path) == len(parent.path)):
                ret.append(newobj)
            else:
                parentpath = path[0 : depth - 1]
                parentobj = lnk[str(parentpath)]
                if "children" not in parentobj:
                    parentobj["children"] = []
                parentobj["children"].append(newobj)
            lnk[str(path)] = newobj
        return ret

    @classmethod
    def get_tree(cls, parent=None):
        """
        :returns:

            A *queryset* of nodes ordered as DFS, including the parent.
            If no parent is given, the entire tree is returned.
        """
        cls = cls.tree_model()

        if parent is None:
            # return the entire tree
            return cls.objects.all()

        return cls.objects.filter(path__descendants=parent.path)

    @classmethod
    def get_root_nodes(cls):
        """:returns: A queryset containing the root nodes in the tree."""
        return cls.tree_model().objects.filter(path__depth=1).order_by("path")

    @classmethod
    def get_descendants_group_count(cls, parent=None):
        """
        Helper for a very common case: get a group of siblings and the number
        of *descendants* (not only children) in every sibling.

        :param parent:

            The parent of the siblings to return. If no parent is given, the
            root nodes will be returned.

        :returns:

            A Queryset of node objects with an extra attribute: `descendants_count`.
        """
        cls = cls.tree_model()
        qs = parent.get_children() if parent else cls.get_root_nodes()
        subquery = (
            cls.objects.filter(path__descendants=OuterRef("path"))
            .order_by()
            .annotate(count=Func(F("pk"), function="Count"))
            .values("count")
        )
        return qs.annotate(
            descendants_count=Subquery(subquery, output_field=models.IntegerField()) - 1
        )  # Subtract the parent node from the count

    def get_depth(self):
        """:returns: the depth (level) of the node"""
        return len(self.path)

    def get_siblings(self):
        """
        :returns: A queryset of all the node's siblings, including the node
            itself.
        """
        return self.tree_model().objects.filter(path__descendants=self.path[:-1], path__depth=len(self.path))

    def get_children(self):
        """:returns: A queryset of all the node's children"""
        return self.tree_model().objects.filter(path__descendants=self.path, path__depth=len(self.path) + 1)

    def get_next_sibling(self):
        """
        :returns: The next node's sibling, or None if it was the rightmost
            sibling.
        """
        return self.get_siblings().filter(path__gt=self.path).first()

    def get_descendants(self, include_self=False):
        """
        :returns: A queryset of all the node's descendants as DFS, doesn't
            include the node itself if `include_self` is False
        """
        if include_self:
            return self.__class__.get_tree(self)

        return self.__class__.get_tree(self).exclude(pk=self.pk)

    def get_prev_sibling(self):
        """
        :returns: The previous node's sibling, or None if it was the leftmost
            sibling.
        """
        try:
            return self.get_siblings().filter(path__lt=self.path).reverse()[0]
        except IndexError:
            return None

    def get_children_count(self):
        """
        :returns: The number the node's children, calculated in the most
        efficient possible way.
        """
        return self.get_children().count()

    def is_sibling_of(self, node) -> bool:
        """
        :returns: ``True`` if the node is a sibling of another node given as an
            argument, else, returns ``False``
        """
        return self.path[:-1] == node.path[:-1]

    def is_child_of(self, node) -> bool:
        """
        :returns: ``True`` is the node if a child of another node given as an
            argument, else, returns ``False``
        """
        return self.path[:-1] == node.path

    def is_descendant_of(self, node) -> bool:
        """
        :returns: ``True`` if the node is a descendant of another node given
            as an argument, else, returns ``False``
        """
        return len(self.path) > len(node.path) and all([self.path[idx] == label for idx, label in enumerate(node.path)])

    @transaction.atomic
    def add_child(self, **kwargs):
        """
        Adds a child to the node.

        This method saves the node in database. The object is populated as if via:

        ```
        obj = self.__class__(**kwargs)
        ```
        """
        return LT_AddChildHandler(self, kwargs).process()

    @transaction.atomic
    def add_sibling(self, pos=None, **kwargs):
        """
        Adds a new node as a sibling to the current node object.

        This method saves the node in database. The object is populated as if via:

        ```
        obj = self.__class__(**kwargs)
        ```
        """
        return LT_AddSiblingHandler(self, pos, kwargs).process()

    def get_root(self):
        """:returns: the root node for the current node object."""
        return self.tree_model().objects.get(path=self.path[0])

    def is_root(self):
        """:returns: True if the node is a root node (else, returns False)"""
        return len(self.path) == 1

    def is_leaf(self):
        """:returns: True if the node is a leaf node (else, returns False)"""
        return not self.get_children().exists()

    def get_ancestors(self):
        """
        :returns: A queryset containing the current node object's ancestors,
            starting by the root node and descending to the parent.
        """
        return self.tree_model().objects.filter(path__ancestors=self.path).exclude(pk=self.pk)

    def get_parent(self, update=False):
        """
        :returns: the parent node of the current node object.
        """
        if len(self.path) == 1:
            return

        parentpath = self.path[:-1]
        return self.tree_model().objects.get(path=parentpath)

    @transaction.atomic
    def move(self, target, pos=None):
        """
        Moves the current node and all it's descendants to a new position
        relative to another node.
        """
        return LT_MoveHandler(self, target, pos).process()

    class Meta:
        abstract = True



"""Pytest configuration file"""

import os

import pytest

os.environ["DJANGO_SETTINGS_MODULE"] = "tests.settings"

import django
from django.apps import apps
from django.db import connection
from django.db.models.signals import pre_migrate


def pytest_report_header(config):
    return "Django: " + django.get_version()


def pytest_configure(config):
    django.setup()


def _pre_migration(*args, **kwargs):
    if os.environ.get("DATABASE_ENGINE", "") == "psql":
        with connection.cursor() as cursor:
            cursor.execute("CREATE EXTENSION IF NOT EXISTS ltree;")


@pytest.fixture(autouse=True, scope="session")
def django_test_environment(django_test_environment):
    pre_migrate.connect(_pre_migration, sender=apps.get_app_config("treebeard"))



"""Adjacency List"""

from django.core import serializers
from django.db import models, transaction
from django.db.models import Max, Min
from django.utils.translation import gettext_noop as _

from treebeard.exceptions import InvalidMoveToDescendant, NodeAlreadySaved
from treebeard.models import Node


class AL_NodeManager(models.Manager):
    """Custom manager for nodes in an Adjacency List tree."""

    def get_queryset(self):
        """Sets the custom queryset as the default."""
        if self.model.node_order_by:
            order_by = ["parent"] + list(self.model.node_order_by)
        else:
            order_by = ["parent", "sib_order"]
        return super().get_queryset().order_by(*order_by)


class AL_Node(Node):
    """Abstract model to create your own Adjacency List Trees."""

    objects = AL_NodeManager()
    node_order_by = None

    TREEBEARD_IDENTIFYING_FIELD = "parent"
    MOVENODE_FORM_EXCLUDED_FIELDS = ("sib_order", "parent")

    _cached_attributes = (
        *Node._cached_attributes,
        "_cached_depth",
    )

    @classmethod
    @transaction.atomic
    def add_root(cls, **kwargs):
        """Adds a root node to the tree."""

        if len(kwargs) == 1 and "instance" in kwargs:
            # adding the passed (unsaved) instance to the tree
            newobj = kwargs["instance"]
            if not newobj._state.adding:
                raise NodeAlreadySaved("Attempted to add a tree node that is already in the database")
        else:
            newobj = cls(**kwargs)

        newobj._cached_depth = 1
        if not cls.node_order_by:
            max = cls.tree_model().objects.filter(parent__isnull=True).aggregate(max=Max("sib_order"))["max"] or 0
            newobj.sib_order = max + 1
        newobj.save()
        return newobj

    @classmethod
    def get_root_nodes(cls):
        """:returns: A queryset containing the root nodes in the tree."""
        return cls.tree_model().objects.filter(parent__isnull=True)

    def get_depth(self, update=False):
        """
        :returns: the depth (level) of the node
            Caches the result in the object itself to help in loops.

        :param update: Updates the cached value.
        """

        if self.parent_id is None:
            return 1

        try:
            if update:
                del self._cached_depth
            else:
                return self._cached_depth
        except AttributeError:
            pass

        depth = 0
        node = self
        while node:
            node = node.parent
            depth += 1
        self._cached_depth = depth
        return depth

    def get_children(self):
        """:returns: A queryset of all the node's children"""
        return self.tree_model().objects.filter(parent=self)

    def get_parent(self, update=False):
        """:returns: the parent node of the current node object."""
        if self._meta.proxy_for_model:
            # the current node is a proxy model; the returned parent
            # should be the same proxy model, so we need to explicitly
            # fetch it as an instance of that model rather than simply
            # following the 'parent' relation
            if self.parent_id is None:
                return None
            else:
                return self.__class__.objects.get(pk=self.parent_id)
        else:
            return self.parent

    def get_ancestors(self):
        """
        :returns: A *list* containing the current node object's ancestors,
            starting by the root node and descending to the parent.
        """
        ancestors = []
        if self._meta.proxy_for_model:
            # the current node is a proxy model; our result set
            # should use the same proxy model, so we need to
            # explicitly fetch instances of that model
            # when following the 'parent' relation
            cls = self.__class__
            node = self
            while node.parent_id:
                node = cls.objects.get(pk=node.parent_id)
                ancestors.insert(0, node)
        else:
            node = self.parent
            while node:
                ancestors.insert(0, node)
                node = node.parent
        return ancestors

    def get_root(self):
        """:returns: the root node for the current node object."""
        ancestors = self.get_ancestors()
        if ancestors:
            return ancestors[0]
        return self

    def is_descendant_of(self, node):
        """
        :returns: ``True`` if the node if a descendant of another node given
            as an argument, else, returns ``False``
        """
        return self.pk in (obj.pk for obj in node.get_descendants())

    @classmethod
    def dump_bulk(cls, parent=None, keep_ids=True):
        """Dumps a tree branch to a python data structure."""

        # a list of nodes: not really a queryset, but it works
        objs = cls.get_tree(parent)

        ret, lnk = [], {}
        pk_field = cls._meta.pk.attname
        for node, pyobj in zip(objs, serializers.serialize("python", objs)):
            depth = node.get_depth()
            # django's serializer stores the attributes in 'fields'
            fields = pyobj["fields"]
            del fields["parent"]

            # non-sorted trees have this
            if "sib_order" in fields:
                del fields["sib_order"]

            if pk_field in fields:
                del fields[pk_field]

            newobj = {"data": fields}
            if keep_ids:
                newobj[pk_field] = pyobj["pk"]

            if (not parent and depth == 1) or (parent and depth == parent.get_depth()):
                ret.append(newobj)
            else:
                parentobj = lnk[node.parent_id]
                if "children" not in parentobj:
                    parentobj["children"] = []
                parentobj["children"].append(newobj)
            lnk[node.pk] = newobj
        return ret

    @transaction.atomic
    def add_child(self, **kwargs):
        """Adds a child to the node."""
        cls = self.tree_model()

        if len(kwargs) == 1 and "instance" in kwargs:
            # adding the passed (unsaved) instance to the tree
            newobj = kwargs["instance"]
            if not newobj._state.adding:
                raise NodeAlreadySaved("Attempted to add a tree node that is already in the database")
        else:
            newobj = cls(**kwargs)

        try:
            newobj._cached_depth = self._cached_depth + 1
        except AttributeError:
            pass
        if not cls.node_order_by:
            max = cls.objects.filter(parent=self).aggregate(max=Max("sib_order"))["max"] or 0
            newobj.sib_order = max + 1
        newobj.parent = self
        newobj.save()
        return newobj

    @classmethod
    def _get_tree_recursively(cls, results, parent, depth):
        if parent:
            nodes = parent.get_children()
        else:
            nodes = cls.get_root_nodes()
        for node in nodes:
            node._cached_depth = depth
            results.append(node)
            cls._get_tree_recursively(results, node, depth + 1)

    @classmethod
    def get_tree(cls, parent=None):
        """
        :returns: A list of nodes ordered as DFS, including the parent. If
                  no parent is given, the entire tree is returned.
        """
        if parent:
            depth = parent.get_depth() + 1
            results = [parent]
        else:
            depth = 1
            results = []
        cls._get_tree_recursively(results, parent, depth)
        return results

    def get_descendants(self, include_self=False):
        """
        :returns: A *list* of all the node's descendants, doesn't
            include the node itself if `include_self` is False
        """
        if include_self:
            return self.__class__.get_tree(self)
        return self.__class__.get_tree(parent=self)[1:]

    def get_descendant_count(self):
        """:returns: the number of descendants of a nodee"""
        return len(self.get_descendants())

    def get_siblings(self):
        """
        :returns: A queryset of all the node's siblings, including the node
            itself.
        """
        if self.parent_id:
            return self.tree_model().objects.filter(parent_id=self.parent_id)
        return self.__class__.get_root_nodes()

    def get_prev_sibling(self):
        return self.get_siblings().filter(sib_order__lt=self.sib_order).last()

    def get_next_sibling(self):
        return self.get_siblings().filter(sib_order__gt=self.sib_order).first()

    @transaction.atomic
    def add_sibling(self, pos=None, **kwargs):
        """Adds a new node as a sibling to the current node object."""
        pos = self._prepare_pos_var_for_add_sibling(pos)

        if len(kwargs) == 1 and "instance" in kwargs:
            # adding the passed (unsaved) instance to the tree
            newobj = kwargs["instance"]
            if not newobj._state.adding:
                raise NodeAlreadySaved("Attempted to add a tree node that is already in the database")
        else:
            # creating a new object
            newobj = self.tree_model()(**kwargs)

        if not self.node_order_by:
            newobj.sib_order = self.__class__._get_new_sibling_order(pos, self)
        newobj.parent_id = self.parent_id
        newobj.save()
        return newobj

    @classmethod
    def _is_target_pos_the_last_sibling(cls, pos, target):
        return pos == "last-sibling" or (pos == "right" and target == target.get_last_sibling())

    @classmethod
    def _make_hole_in_db(cls, min, target_node):
        qset = cls.tree_model().objects.filter(sib_order__gte=min)
        if target_node.is_root():
            qset = qset.filter(parent__isnull=True)
        else:
            qset = qset.filter(parent=target_node.parent)
        qset.update(sib_order=models.F("sib_order") + 1)

    @classmethod
    def _make_hole_and_get_sibling_order(cls, pos, target_node):
        siblings = target_node.get_siblings()
        siblings = {
            "left": siblings.filter(sib_order__gte=target_node.sib_order),
            "right": siblings.filter(sib_order__gt=target_node.sib_order),
            "first-sibling": siblings,
        }[pos]
        sib_order = {"left": target_node.sib_order, "right": target_node.sib_order + 1, "first-sibling": 1}[pos]
        min = siblings.aggregate(min=Min("sib_order"))["min"] or 0
        if min:
            cls._make_hole_in_db(min, target_node)
        return sib_order

    @classmethod
    def _get_new_sibling_order(cls, pos, target_node):
        if cls._is_target_pos_the_last_sibling(pos, target_node):
            sib_order = target_node.get_last_sibling().sib_order + 1
        else:
            sib_order = cls._make_hole_and_get_sibling_order(pos, target_node)
        return sib_order

    @transaction.atomic
    def move(self, target, pos=None):
        """
        Moves the current node and all it's descendants to a new position
        relative to another node.
        """

        pos = self._prepare_pos_var_for_move(pos)

        sib_order = None
        parent = None

        if pos in ("first-child", "last-child", "sorted-child"):
            if self == target:
                raise InvalidMoveToDescendant(_("Can't move node to itself."))

            # moving to a child
            if not target.is_leaf():
                target = target.get_last_child()
                pos = {"first-child": "first-sibling", "last-child": "last-sibling", "sorted-child": "sorted-sibling"}[
                    pos
                ]
            else:
                parent = target
                if pos == "sorted-child":
                    pos = "sorted-sibling"
                else:
                    pos = "first-sibling"
                    sib_order = 1

        if target.is_descendant_of(self):
            raise InvalidMoveToDescendant(_("Can't move node to a descendant."))

        if self == target and (
            (pos == "left")
            or (pos in ("right", "last-sibling") and target == target.get_last_sibling())
            or (pos == "first-sibling" and target == target.get_first_sibling())
        ):
            # special cases, not actually moving the node so no need to UPDATE
            return

        if pos == "sorted-sibling":
            if parent:
                self.parent = parent
            else:
                self.parent = target.parent
        else:
            if sib_order:
                self.sib_order = sib_order
            else:
                self.sib_order = self.__class__._get_new_sibling_order(pos, target)
            if parent:
                self.parent = parent
            else:
                self.parent = target.parent

        self.save()

    class Meta:
        """Abstract model."""

        abstract = True



"""
See PEP 386 (https://www.python.org/dev/peps/pep-0386/)

Release logic:
 1. Remove ".devX" from __version__ (below)
 2. git add treebeard/__init__.py
 3. git commit -m 'Bump to <version>'
 4. git tag <version>
 5. git push
 6. ensure that all tests pass on Github Actions
 7. git push --tags
 8. pip install --upgrade pip build twine
 9. python -m build
10. twine upload dist/*
11. bump the version, append ".dev0" to __version__
12. git add treebeard/__init__.py
13. git commit -m 'Start with <version>'
14. git push
"""

__version__ = "5.0.5"



from django.contrib.admin.templatetags.admin_list import result_list
from django.template import Library

register = Library()


def _get_parent_id(node):
    """Return the node's parent id or 0 if node is a root node."""
    if node.is_root():
        return 0
    return node.get_parent().pk


@register.inclusion_tag("admin/change_list_results.html")
def result_tree(cl):
    return result_list(cl)


@register.simple_tag
def tree_context(cl):
    """
    Generate a list containing additional context for each row in the list, for use
    by the frontend. It is assumed that the template renders the items in the same order
    as the `result_list` iterator.
    """
    return [
        {
            "node-id": str(obj.pk),
            "parent-id": _get_parent_id(obj),
            "level": obj.get_depth(),
            "has-children": int(not obj.is_leaf()),
        }
        for obj in cl.result_list
    ]



from collections import UserList
from collections.abc import Iterable

from django import forms
from django.core.validators import RegexValidator
from django.db.models import CharField, IntegerField
from django.db.models.expressions import Func
from django.db.models.fields import TextField
from django.db.models.lookups import PostgresOperatorLookup, Transform
from django.forms.widgets import TextInput

# PathField implementation borrows significantly from https://github.com/mariocesar/django-ltree/blob/master/django_ltree/fields.py

path_label_validator = RegexValidator(
    r"^(?P<root>[a-zA-Z0-9_-]+)(?:\.[a-zA-Z0-9_-]+)*$",
    "A label is a sequence of alphanumeric characters and underscores separated by dots or slashes.",
    "invalid",
)


class PathValue(UserList):
    def __init__(self, value):
        if isinstance(value, str):
            value = value.strip().split(".") if value else []
        elif isinstance(value, Iterable):
            value = list(value)
        else:
            raise ValueError(f"Invalid value: {value!r} for path")

        super().__init__(initlist=value)

    def __repr__(self):
        return str(self)

    def __str__(self):
        return ".".join(self)


class PathValueProxy:
    def __init__(self, field_name):
        self.field_name = field_name

    def __get__(self, instance, owner):
        if instance is None:
            return self

        value = instance.__dict__[self.field_name]

        if value is None:
            return value

        return PathValue(instance.__dict__[self.field_name])

    def __set__(self, instance, value):
        if instance is None:
            return self

        instance.__dict__[self.field_name] = value


class PathFormField(forms.CharField):
    default_validators = [path_label_validator]


class PathField(TextField):
    default_validators = [path_label_validator]

    def db_type(self, connection):
        return "ltree"

    def formfield(self, **kwargs):
        kwargs["form_class"] = PathFormField
        kwargs["widget"] = TextInput(attrs={"class": "vTextField"})
        return super().formfield(**kwargs)

    def contribute_to_class(self, cls, name, private_only=False):
        super().contribute_to_class(cls, name)
        setattr(cls, self.name, PathValueProxy(self.name))

    def from_db_value(self, value, expression, connection, *args):
        if value is None:
            return value
        return PathValue(value)

    def get_prep_value(self, value):
        if value is None:
            return value
        return str(PathValue(value))

    def to_python(self, value):
        if value is None:
            return value

        if isinstance(value, PathValue):
            return value

        return PathValue(value)

    def get_db_prep_value(self, value, connection, prepared=False):
        if value is None:
            return value

        if isinstance(value, PathValue):
            return str(value)

        if isinstance(value, (list, str)):
            return str(PathValue(value))

        raise ValueError(f"Unknown value type {type(value)}")


@PathField.register_lookup
class AncestorLookup(PostgresOperatorLookup):
    lookup_name = "ancestors"
    postgres_operator = "@>"


@PathField.register_lookup
class DescendantLookup(PostgresOperatorLookup):
    lookup_name = "descendants"
    postgres_operator = "<@"


class Subpath(Func):
    function = "subpath"
    output_field = PathField()

    def __init__(self, expression, pos, length=None, **extra):
        """
        expression: the name of a field, or an expression returning a string
        pos: an integer >= 0, or an expression returning an integer
        length: an optional number of labels to return
        """
        if not hasattr(pos, "resolve_expression"):
            if pos < 0:
                raise ValueError("'pos' must be positive")
        expressions = [expression, pos]
        if length is not None:
            expressions.append(length)
        super().__init__(*expressions, **extra)


class Ltree2Text(Transform):
    function = "ltree2text"
    lookup_name = "ltree2text"
    output_field = CharField()


class Text2LTree(Transform):
    function = "text2ltree"
    lookup_name = "text2ltree"
    output_field = PathField()


@PathField.register_lookup
class NLevel(Transform):
    lookup_name = "depth"
    function = "nlevel"
    output_field = IntegerField()



"""
Django settings for testing treebeard
"""

import os


def get_db_conf():
    """
    Configures database according to the DATABASE_ENGINE environment
    variable. Defaults to SQlite.

    This method is used to run tests against different database backends.
    """
    database_engine = os.environ.get("DATABASE_ENGINE", "sqlite")
    if database_engine == "sqlite":
        return {"ENGINE": "django.db.backends.sqlite3", "NAME": ":memory:"}
    elif database_engine == "psql":
        return {
            "ENGINE": "django.db.backends.postgresql",
            "NAME": "treebeard",
            "USER": os.environ.get("DATABASE_USER_POSTGRES", "treebeard"),
            "PASSWORD": os.environ.get("DATABASE_PASSWORD", ""),
            "HOST": os.environ.get("DATABASE_HOST", "localhost"),
            "PORT": os.environ.get("DATABASE_PORT_POSTGRES", ""),
        }
    elif database_engine == "mysql":
        return {
            "ENGINE": "django.db.backends.mysql",
            "NAME": "treebeard",
            "USER": os.environ.get("DATABASE_USER_MYSQL", "treebeard"),
            "PASSWORD": os.environ.get("DATABASE_PASSWORD", ""),
            "HOST": os.environ.get("DATABASE_HOST", "localhost"),
            "PORT": os.environ.get("DATABASE_PORT_MYSQL", ""),
        }
    elif database_engine == "mssql":
        return {
            "ENGINE": "mssql",
            "NAME": "master",
            "USER": "sa",
            "PASSWORD": "Password12!",
            "HOST": "localhost",
            "PORT": os.environ.get("DATABASE_PORT_MSSQL", ""),
            "OPTIONS": {
                "driver": "ODBC Driver 18 for SQL Server",
                "extra_params": "Trusted_Connection=no;TrustServerCertificate=yes",
            },
        }


DATABASES = {"default": get_db_conf()}
SECRET_KEY = "7r33b34rd"

INSTALLED_APPS = [
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.admin",
    "django.contrib.messages",
    "treebeard",
    "tests",
]

MIDDLEWARE = [
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
]

ROOT_URLCONF = "tests.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.contrib.auth.context_processors.auth",
                "django.template.context_processors.debug",
                "django.template.context_processors.i18n",
                "django.template.context_processors.media",
                "django.template.context_processors.static",
                "django.template.context_processors.tz",
                "django.template.context_processors.request",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"



import time

import pytest

from treebeard.ltree import InvalidLabelConstraints, generate_label


class TestGenerateLabel:
    def test_invalid_before_after(self):
        with pytest.raises(InvalidLabelConstraints):
            # Before is less than after
            generate_label(before="A", after="B", skip=set())

        with pytest.raises(InvalidLabelConstraints):
            # Before is equal to after
            generate_label(before="A", after="A", skip=set())

    def test_no_constraint(self):
        assert generate_label(skip=set()) == "A"

    def test_no_after(self):
        assert generate_label(before="A", skip=set()) == "0"
        assert generate_label(before="ABCDE", skip=set()) == "0"

    def test_no_before(self):
        assert generate_label(after="A", skip=set()) == "B"

    def test_reduces_char_count_where_possible(self):
        assert generate_label(after="ABB", skip=set()) == "B"
        assert generate_label(after="ZYX", skip=set()) == "ZZ"
        assert generate_label(after="ZYX", before="ZZ", skip=set()) == "ZYY"

    def test_adds_char_if_needed(self):
        assert generate_label(before="AA", after="A", skip=set()) == "A0"
        assert generate_label(before="B", after="AZ", skip=set()) == "AZ0"

    def test_respects_skip(self):
        assert generate_label(before="AA", after="A", skip={"A0"}) == "A1"
        assert generate_label(before="B", after="AZ", skip={"AZ0"}) == "AZ1"
        assert generate_label(after="ZYX", skip={"ZZ"}) == "ZYY"

    def test_large_string(self):
        # Check that we don't have exponential time for large strings
        start = time.perf_counter()
        res = generate_label(before="A" * 59999 + "B", after="A" * 60000, skip=set())
        time_taken = time.perf_counter() - start
        assert time_taken < 0.5  # Conservative to allow for flakiness on CI
        assert res == "A" * 60000 + "0"

        start = time.perf_counter()
        res = generate_label(before="B", after="A" * 60000, skip=set())
        time_taken = time.perf_counter() - start
        assert time_taken < 0.05
        assert res == "AB"



# Generated by Django 3.1.2 on 2021-02-24 20:44

import os
import uuid

import django.db.models.deletion
from django.db import migrations, models

import treebeard.ltree.fields


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="AL_TestNode",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("sib_order", models.PositiveIntegerField()),
                ("desc", models.CharField(max_length=255)),
                (
                    "parent",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="children_set",
                        to="tests.al_testnode",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="MP_TestNode",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("path", models.CharField(max_length=255, unique=True)),
                ("depth", models.PositiveIntegerField()),
                ("numchild", models.PositiveIntegerField(default=0)),
                ("desc", models.CharField(max_length=255)),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="MP_TestNodeAlphabet",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("path", models.CharField(max_length=255, unique=True)),
                ("depth", models.PositiveIntegerField()),
                ("numchild", models.PositiveIntegerField(default=0)),
                ("numval", models.IntegerField()),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="MP_TestNodeCustomId",
            fields=[
                ("path", models.CharField(max_length=255, unique=True)),
                ("depth", models.PositiveIntegerField()),
                ("numchild", models.PositiveIntegerField(default=0)),
                ("id", models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ("desc", models.CharField(max_length=255)),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="MP_TestNodeShortPath",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("path", models.CharField(max_length=4, unique=True)),
                ("depth", models.PositiveIntegerField()),
                ("numchild", models.PositiveIntegerField(default=0)),
                ("desc", models.CharField(max_length=255)),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="MP_TestNodeSmallStep",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("path", models.CharField(max_length=255, unique=True)),
                ("depth", models.PositiveIntegerField()),
                ("numchild", models.PositiveIntegerField(default=0)),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="MP_TestNodeSorted",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("path", models.CharField(max_length=255, unique=True)),
                ("depth", models.PositiveIntegerField()),
                ("numchild", models.PositiveIntegerField(default=0)),
                ("val1", models.IntegerField()),
                ("val2", models.IntegerField()),
                ("desc", models.CharField(max_length=255)),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="MP_TestNodeInheritedSorted",
            fields=[
                (
                    "mp_testnodesorted_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="tests.mp_testnodesorted",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=("tests.mp_testnodesorted",),
        ),
        migrations.CreateModel(
            name="MP_TestNodeSortedAutoNow",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("path", models.CharField(max_length=255, unique=True)),
                ("depth", models.PositiveIntegerField()),
                ("numchild", models.PositiveIntegerField(default=0)),
                ("desc", models.CharField(max_length=255)),
                ("created", models.DateTimeField(auto_now_add=True)),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="MP_TestNodeUuid",
            fields=[
                ("path", models.CharField(max_length=255, unique=True)),
                ("depth", models.PositiveIntegerField()),
                ("numchild", models.PositiveIntegerField(default=0)),
                ("custom_id", models.UUIDField(default=uuid.uuid1, editable=False, primary_key=True, serialize=False)),
                ("desc", models.CharField(max_length=255)),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="MP_TestSortedNodeShortPath",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("path", models.CharField(max_length=4, unique=True)),
                ("depth", models.PositiveIntegerField()),
                ("numchild", models.PositiveIntegerField(default=0)),
                ("desc", models.CharField(max_length=255)),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="NS_TestNode",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("lft", models.PositiveIntegerField(db_index=True)),
                ("rgt", models.PositiveIntegerField(db_index=True)),
                ("tree_id", models.PositiveIntegerField(db_index=True)),
                ("depth", models.PositiveIntegerField(db_index=True)),
                ("desc", models.CharField(max_length=255)),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="NS_TestNodeSorted",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("lft", models.PositiveIntegerField(db_index=True)),
                ("rgt", models.PositiveIntegerField(db_index=True)),
                ("tree_id", models.PositiveIntegerField(db_index=True)),
                ("depth", models.PositiveIntegerField(db_index=True)),
                ("val1", models.IntegerField()),
                ("val2", models.IntegerField()),
                ("desc", models.CharField(max_length=255)),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="RelatedModel",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("desc", models.CharField(max_length=255)),
            ],
        ),
        migrations.CreateModel(
            name="AL_TestNodeInherited",
            fields=[
                (
                    "al_testnode_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="tests.al_testnode",
                    ),
                ),
                ("extra_desc", models.CharField(max_length=255)),
            ],
            options={
                "abstract": False,
            },
            bases=("tests.al_testnode",),
        ),
        migrations.CreateModel(
            name="MP_TestNodeInherited",
            fields=[
                (
                    "mp_testnode_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="tests.mp_testnode",
                    ),
                ),
                ("extra_desc", models.CharField(max_length=255)),
            ],
            options={
                "abstract": False,
            },
            bases=("tests.mp_testnode",),
        ),
        migrations.CreateModel(
            name="NS_TestNodeInherited",
            fields=[
                (
                    "ns_testnode_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="tests.ns_testnode",
                    ),
                ),
                ("extra_desc", models.CharField(max_length=255)),
            ],
            options={
                "abstract": False,
            },
            bases=("tests.ns_testnode",),
        ),
        migrations.CreateModel(
            name="NS_TestNodeSomeDep",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("node", models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="tests.ns_testnode")),
            ],
        ),
        migrations.CreateModel(
            name="NS_TestNodeRelated",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("lft", models.PositiveIntegerField(db_index=True)),
                ("rgt", models.PositiveIntegerField(db_index=True)),
                ("tree_id", models.PositiveIntegerField(db_index=True)),
                ("depth", models.PositiveIntegerField(db_index=True)),
                ("desc", models.CharField(max_length=255)),
                ("related", models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="tests.relatedmodel")),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="MP_TestNodeSomeDep",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("node", models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="tests.mp_testnode")),
            ],
        ),
        migrations.CreateModel(
            name="MP_TestNodeRelated",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("path", models.CharField(max_length=255, unique=True)),
                ("depth", models.PositiveIntegerField()),
                ("numchild", models.PositiveIntegerField(default=0)),
                ("desc", models.CharField(max_length=255)),
                ("related", models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="tests.relatedmodel")),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="AL_TestNodeSorted",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("val1", models.IntegerField()),
                ("val2", models.IntegerField()),
                ("desc", models.CharField(max_length=255)),
                (
                    "parent",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="children_set",
                        to="tests.al_testnodesorted",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="AL_TestNodeSomeDep",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("node", models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="tests.al_testnode")),
            ],
        ),
        migrations.CreateModel(
            name="AL_TestNodeRelated",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("sib_order", models.PositiveIntegerField()),
                ("desc", models.CharField(max_length=255)),
                (
                    "parent",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="children_set",
                        to="tests.al_testnoderelated",
                    ),
                ),
                ("related", models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="tests.relatedmodel")),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="AL_TestNode_Proxy",
            fields=[],
            options={
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=("tests.al_testnode",),
        ),
        migrations.CreateModel(
            name="MP_TestNode_Proxy",
            fields=[],
            options={
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=("tests.mp_testnode",),
        ),
        migrations.CreateModel(
            name="NS_TestNode_Proxy",
            fields=[],
            options={
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=("tests.ns_testnode",),
        ),
    ]


if os.environ.get("DATABASE_ENGINE") == "psql":
    Migration.operations.extend(
        [
            migrations.CreateModel(
                name="LT_TestNode",
                fields=[
                    (
                        "id",
                        models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID"),
                    ),
                    ("path", treebeard.ltree.fields.PathField(unique=True)),
                    ("desc", models.CharField(max_length=255)),
                ],
                options={
                    "abstract": False,
                },
            ),
            migrations.CreateModel(
                name="LT_TestNodeInherited",
                fields=[
                    (
                        "lt_testnode_ptr",
                        models.OneToOneField(
                            auto_created=True,
                            on_delete=django.db.models.deletion.CASCADE,
                            parent_link=True,
                            primary_key=True,
                            serialize=False,
                            to="tests.lt_testnode",
                        ),
                    ),
                    ("extra_desc", models.CharField(max_length=255)),
                ],
                options={
                    "abstract": False,
                },
                bases=("tests.lt_testnode",),
            ),
            migrations.CreateModel(
                name="LT_TestNodeSomeDep",
                fields=[
                    (
                        "id",
                        models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID"),
                    ),
                    ("node", models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="tests.lt_testnode")),
                ],
            ),
            migrations.CreateModel(
                name="LT_TestNode_Proxy",
                fields=[],
                options={
                    "proxy": True,
                    "indexes": [],
                    "constraints": [],
                },
                bases=("tests.lt_testnode",),
            ),
            migrations.CreateModel(
                name="LT_TestNodeSorted",
                fields=[
                    (
                        "id",
                        models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID"),
                    ),
                    ("path", treebeard.ltree.fields.PathField(unique=True)),
                    ("val1", models.IntegerField()),
                    ("val2", models.IntegerField()),
                    ("desc", models.CharField(max_length=255)),
                ],
                options={
                    "abstract": False,
                },
            ),
            migrations.CreateModel(
                name="LT_TestNodeInheritedSorted",
                fields=[
                    (
                        "lt_testnodesorted_ptr",
                        models.OneToOneField(
                            auto_created=True,
                            on_delete=django.db.models.deletion.CASCADE,
                            parent_link=True,
                            primary_key=True,
                            serialize=False,
                            to="tests.lt_testnodesorted",
                        ),
                    ),
                ],
                options={
                    "abstract": False,
                },
                bases=("tests.lt_testnodesorted",),
            ),
        ]
    )



import string

import pytest

from treebeard.numconv import NumConv


@pytest.mark.parametrize(
    ("num", "chars"),
    [
        (0, "0"),
        (1, "1"),
        (26, "Q"),
        (27, "R"),
        (46, "1A"),
        (999, "RR"),
        (10560, "85C"),
    ],
)
def test_numconv(num, chars):
    conv = NumConv(string.digits + string.ascii_uppercase)
    assert conv.int2str(num) == chars
    assert conv.str2int(chars) == num



"""Nested Sets"""

import operator
from functools import reduce

from django.core import serializers
from django.db import models, transaction
from django.db.models import Case, F, Q, When
from django.utils.translation import gettext_noop as _

from treebeard.exceptions import InvalidMoveToDescendant, NodeAlreadySaved
from treebeard.models import Node


def merge_deleted_counters(c1, c2):
    """
    Merge return values from Django's Queryset.delete() method.
    """
    object_counts = {key: c1[1].get(key, 0) + c2[1].get(key, 0) for key in set(c1[1]) | set(c2[1])}
    return (c1[0] + c2[0], object_counts)


class NS_NodeQuerySet(models.query.QuerySet):
    """
    Custom queryset for the tree node manager.

    Needed only for the customized delete method.
    """

    def delete(self, *args, removed_ranges=None, deleted_counter=None, **kwargs):
        """
        Custom delete method, will remove all descendant nodes to ensure a
        consistent tree (no orphans)

        :returns: tuple of the number of objects deleted and a dictionary
                  with the number of deletions per object type
        """
        model = self.model.tree_model()

        if deleted_counter is None:
            deleted_counter = (0, {})

        if removed_ranges is not None:
            # we already know the children, let's call the default django
            # delete method and let it handle the removal of the user's
            # foreign keys...
            result = super().delete(*args, **kwargs)
            deleted_counter = merge_deleted_counters(deleted_counter, result)

            # Now closing the gap (Celko's trees book, page 62)
            # We do this for every gap that was left in the tree when the nodes
            # were removed.  If many nodes were removed, we're going to update
            # the same nodes over and over again. This would be probably
            # cheaper precalculating the gapsize per intervals, or just do a
            # complete reordering of the tree (uses COUNT)...
            for tree_id, drop_lft, drop_rgt in sorted(removed_ranges, reverse=True):
                model._close_gap(drop_lft, drop_rgt, tree_id)
        else:
            # we'll have to manually run through all the nodes that are going
            # to be deleted and remove nodes from the list if an ancestor is
            # already getting removed, since that would be redundant
            removed = {}
            for node in self.order_by("tree_id", "lft"):
                found = False
                for rid, rnode in removed.items():
                    if node.is_descendant_of(rnode):
                        found = True
                        break
                if not found:
                    removed[node.pk] = node

            # ok, got the minimal list of nodes to remove...
            # we must also remove their descendants
            toremove = []
            ranges = []
            for id, node in removed.items():
                toremove.append(Q(lft__range=(node.lft, node.rgt)) & Q(tree_id=node.tree_id))
                ranges.append((node.tree_id, node.lft, node.rgt))
            if toremove:
                deleted_counter = model.objects.filter(reduce(operator.or_, toremove)).delete(
                    removed_ranges=ranges, deleted_counter=deleted_counter
                )
        return deleted_counter

    delete.alters_data = True
    delete.queryset_only = True


class NS_NodeManager(models.Manager):
    """Custom manager for nodes in a Nested Sets tree."""

    def get_queryset(self):
        """Sets the custom queryset as the default."""
        return NS_NodeQuerySet(self.model).order_by("tree_id", "lft")


class NS_Node(Node):
    """Abstract model to create your own Nested Sets Trees."""

    node_order_by = []

    lft = models.PositiveIntegerField(db_index=True)
    rgt = models.PositiveIntegerField(db_index=True)
    tree_id = models.PositiveIntegerField(db_index=True)
    depth = models.PositiveIntegerField(db_index=True)

    TREEBEARD_IDENTIFYING_FIELD = "lft"
    MOVENODE_FORM_EXCLUDED_FIELDS = ("depth", "lft", "rgt", "tree_id")

    objects = NS_NodeManager()

    _cached_attributes = (
        *Node._cached_attributes,
        "_cached_parent_obj",
    )

    @classmethod
    @transaction.atomic
    def add_root(cls, **kwargs):
        """Adds a root node to the tree."""

        # do we have a root node already?
        last_root = cls.get_last_root_node()

        if last_root and last_root.node_order_by:
            # there are root nodes and node_order_by has been set
            # delegate sorted insertion to add_sibling
            return last_root.add_sibling("sorted-sibling", **kwargs)

        if last_root:
            # adding the new root node as the last one
            newtree_id = last_root.tree_id + 1
        else:
            # adding the first root node
            newtree_id = 1

        if len(kwargs) == 1 and "instance" in kwargs:
            # adding the passed (unsaved) instance to the tree
            newobj = kwargs["instance"]
            if not newobj._state.adding:
                raise NodeAlreadySaved("Attempted to add a tree node that is already in the database")
        else:
            # creating the new object
            newobj = cls.tree_model()(**kwargs)

        newobj.depth = 1
        newobj.tree_id = newtree_id
        newobj.lft = 1
        newobj.rgt = 2
        # saving the instance before returning it
        newobj.save()
        return newobj

    @classmethod
    def _move_right(cls, tree_id, rgt, lftmove=False, incdec=2):
        lftop = "lft__gte" if lftmove else "lft__gt"
        output_field = models.PositiveIntegerField()
        cls.objects.filter(rgt__gte=rgt, tree_id=tree_id).update(
            lft=Case(When(**{lftop: rgt}, then=F("lft") + incdec), default=F("lft"), output_field=output_field),
            rgt=Case(When(rgt__gte=rgt, then=F("rgt") + incdec), default=F("rgt"), output_field=output_field),
        )

    @classmethod
    def _move_tree_right(cls, tree_id):
        cls.objects.filter(tree_id__gte=tree_id).update(tree_id=F("tree_id") + 1)

    @transaction.atomic
    def add_child(self, **kwargs):
        """Adds a child to the node."""
        # Fetch the parent afresh from the database and lock the row
        # This guards against race conditions and state drift when adding multiple children
        node = self.__class__.objects.filter(pk=self.pk).select_for_update().get()
        if not node.is_leaf():
            # there are child nodes, delegate insertion to add_sibling
            if self.node_order_by:
                pos = "sorted-sibling"
            else:
                pos = "last-sibling"
            last_child = node.get_last_child()
            last_child._cached_parent_obj = node
            new_sibling = last_child.add_sibling(pos, **kwargs)
            node.rgt += 2
            self.rgt = node.rgt + 2  # Update the rgt of the parent object, which may be used again in a loop
            return new_sibling

        # we're adding the first child of this node
        node.__class__._move_right(node.tree_id, node.rgt, False, 2)

        if len(kwargs) == 1 and "instance" in kwargs:
            # adding the passed (unsaved) instance to the tree
            newobj = kwargs["instance"]
            if not newobj._state.adding:
                raise NodeAlreadySaved("Attempted to add a tree node that is already in the database")
        else:
            # creating a new object
            newobj = node.tree_model()(**kwargs)

        newobj.tree_id = node.tree_id
        newobj.depth = node.depth + 1
        newobj.lft = node.lft + 1
        newobj.rgt = node.lft + 2

        # this is just to update the cache
        self.rgt = node.rgt + 2

        newobj._cached_parent_obj = self

        # saving the instance before returning it
        newobj.save()

        return newobj

    @transaction.atomic
    def add_sibling(self, pos=None, **kwargs):
        """Adds a new node as a sibling to the current node object."""

        pos = self._prepare_pos_var_for_add_sibling(pos)

        if len(kwargs) == 1 and "instance" in kwargs:
            # adding the passed (unsaved) instance to the tree
            newobj = kwargs["instance"]
            if not newobj._state.adding:
                raise NodeAlreadySaved("Attempted to add a tree node that is already in the database")
        else:
            # creating a new object
            newobj = self.tree_model()(**kwargs)

        newobj.depth = self.depth

        target = self

        if target.is_root():
            newobj.lft = 1
            newobj.rgt = 2
            if pos == "sorted-sibling":
                siblings = list(target.get_sorted_pos_queryset(target.get_siblings(), newobj))
                if siblings:
                    pos = "left"
                    target = siblings[0]
                else:
                    pos = "last-sibling"

            last_root = target.__class__.get_last_root_node()
            if (pos == "last-sibling") or (pos == "right" and target == last_root):
                newobj.tree_id = last_root.tree_id + 1
            else:
                newpos = {"first-sibling": 1, "left": target.tree_id, "right": target.tree_id + 1}[pos]
                target.__class__._move_tree_right(newpos)

                newobj.tree_id = newpos
        else:
            newobj.tree_id = target.tree_id

            if pos == "sorted-sibling":
                siblings = list(target.get_sorted_pos_queryset(target.get_siblings(), newobj))
                if siblings:
                    pos = "left"
                    target = siblings[0]
                else:
                    pos = "last-sibling"

            if pos in ("left", "right", "first-sibling"):
                siblings = list(target.get_siblings())

                if pos == "right":
                    if target == siblings[-1]:
                        pos = "last-sibling"
                    else:
                        pos = "left"
                        found = False
                        for node in siblings:
                            if found:
                                target = node
                                break
                            elif node == target:
                                found = True
                if pos == "left":
                    if target == siblings[0]:
                        pos = "first-sibling"
                if pos == "first-sibling":
                    target = siblings[0]

            move_right = self.__class__._move_right

            if pos == "last-sibling":
                newpos = target.get_parent().rgt
                move_right(target.tree_id, newpos, False, 2)
            elif pos == "first-sibling":
                newpos = target.lft
                move_right(target.tree_id, newpos - 1, False, 2)
            elif pos == "left":
                newpos = target.lft
                move_right(target.tree_id, newpos, True, 2)

            newobj.lft = newpos
            newobj.rgt = newpos + 1

        # saving the instance before returning it
        newobj.save()

        return newobj

    @transaction.atomic
    def move(self, target, pos=None):
        """
        Moves the current node and all it's descendants to a new position
        relative to another node.
        """

        pos = self._prepare_pos_var_for_move(pos)
        cls = self.tree_model()
        original_target = target

        parent = None

        if pos in ("first-child", "last-child", "sorted-child"):
            if self == target:
                raise InvalidMoveToDescendant(_("Can't move node to itself."))

            # moving to a child
            if target.is_leaf():
                parent = target
                pos = "last-child"
            else:
                target = target.get_last_child()
                pos = {"first-child": "first-sibling", "last-child": "last-sibling", "sorted-child": "sorted-sibling"}[
                    pos
                ]

        if target.is_descendant_of(self):
            raise InvalidMoveToDescendant(_("Can't move node to a descendant."))

        if self == target and (
            (pos == "left")
            or (pos in ("right", "last-sibling") and target == target.get_last_sibling())
            or (pos == "first-sibling" and target == target.get_first_sibling())
        ):
            # special cases, not actually moving the node so nothing to do
            return

        if pos == "sorted-sibling":
            siblings = list(target.get_sorted_pos_queryset(target.get_siblings(), self))
            if siblings:
                pos = "left"
                target = siblings[0]
            else:
                pos = "last-sibling"
        if pos in ("left", "right", "first-sibling"):
            siblings = list(target.get_siblings())

            if pos == "right":
                if target == siblings[-1]:
                    pos = "last-sibling"
                else:
                    pos = "left"
                    found = False
                    for node in siblings:
                        if found:
                            target = node
                            break
                        elif node == target:
                            found = True
            if pos == "left":
                if target == siblings[0]:
                    pos = "first-sibling"
            if pos == "first-sibling":
                target = siblings[0]

        # ok let's move this
        gap = self.rgt - self.lft + 1
        target_tree = target.tree_id

        # first make a hole
        if pos == "last-child":
            newpos = parent.rgt
            cls._move_right(target.tree_id, newpos, False, gap)
        elif target.is_root():
            newpos = 1
            if pos == "last-sibling":
                target_tree = target.get_siblings().reverse()[0].tree_id + 1
            elif pos == "first-sibling":
                target_tree = 1
                cls._move_tree_right(1)
            elif pos == "left":
                cls._move_tree_right(target.tree_id)
        else:
            if pos == "last-sibling":
                newpos = target.get_parent().rgt
                cls._move_right(target.tree_id, newpos, False, gap)
            elif pos == "first-sibling":
                newpos = target.lft
                cls._move_right(target.tree_id, newpos - 1, False, gap)
            elif pos == "left":
                newpos = target.lft
                cls._move_right(target.tree_id, newpos, True, gap)

        # we refresh 'self' because lft/rgt may have changed
        self.refresh_from_db()

        depthdiff = target.depth - self.depth
        if parent:
            depthdiff += 1

        # move the tree to the hole
        jump = newpos - self.lft
        cls.objects.filter(tree_id=self.tree_id, lft__range=(self.lft, self.rgt)).update(
            tree_id=target_tree,
            lft=F("lft") + jump,
            rgt=F("rgt") + jump,
            depth=F("depth") + depthdiff,
        )

        # close the gap
        cls._close_gap(self.lft, self.rgt, self.tree_id)
        self.refresh_from_db()  # Tree params will have changed
        original_target.refresh_from_db()

    @classmethod
    def _close_gap(cls, drop_lft, drop_rgt, tree_id):
        gapsize = drop_rgt - drop_lft + 1
        output_field = models.PositiveIntegerField()
        cls.objects.filter(Q(tree_id=tree_id) & (Q(lft__gt=drop_lft) | Q(rgt__gt=drop_lft))).update(
            lft=Case(When(lft__gt=drop_lft, then=F("lft") - gapsize), default=F("lft"), output_field=output_field),
            rgt=Case(When(rgt__gt=drop_lft, then=F("rgt") - gapsize), default=F("rgt"), output_field=output_field),
        )

    def get_children(self):
        """:returns: A queryset of all the node's children"""
        return self.get_descendants().filter(depth=self.depth + 1)

    def get_depth(self):
        """:returns: the depth (level) of the node"""
        return self.depth

    def is_leaf(self):
        """:returns: True if the node is a leaf node (else, returns False)"""
        return self.rgt - self.lft == 1

    def get_root(self):
        """:returns: the root node for the current node object."""
        if self.lft == 1:
            return self
        return self.tree_model().objects.get(tree_id=self.tree_id, lft=1)

    def is_root(self):
        """:returns: True if the node is a root node (else, returns False)"""
        return self.lft == 1

    def get_siblings(self):
        """
        :returns: A queryset of all the node's siblings, including the node
            itself.
        """
        if self.lft == 1:
            return self.get_root_nodes()
        return self.get_parent(True).get_children()

    @classmethod
    def dump_bulk(cls, parent=None, keep_ids=True):
        """Dumps a tree branch to a python data structure."""
        qset = cls.get_tree(parent)
        ret, lnk = [], {}
        pk_field = cls._meta.pk.attname
        for pyobj in qset.iterator():
            serobj = serializers.serialize("python", [pyobj])[0]
            # django's serializer stores the attributes in 'fields'
            fields = serobj["fields"]
            depth = fields["depth"]
            # this will be useless in load_bulk
            del fields["lft"]
            del fields["rgt"]
            del fields["depth"]
            del fields["tree_id"]
            if pk_field in fields:
                # this happens immediately after a load_bulk
                del fields[pk_field]

            newobj = {"data": fields}
            if keep_ids:
                newobj[pk_field] = serobj["pk"]

            if (not parent and depth == 1) or (parent and depth == parent.depth):
                ret.append(newobj)
            else:
                parentobj = pyobj.get_parent()
                parentser = lnk[parentobj.pk]
                if "children" not in parentser:
                    parentser["children"] = []
                parentser["children"].append(newobj)
            lnk[pyobj.pk] = newobj
        return ret

    @classmethod
    def get_tree(cls, parent=None):
        """
        :returns:

            A *queryset* of nodes ordered as DFS, including the parent.
            If no parent is given, all trees are returned.
        """
        cls = cls.tree_model()

        if parent is None:
            # return the entire tree
            return cls.objects.all()
        if parent.is_leaf():
            return cls.objects.filter(pk=parent.pk)
        return cls.objects.filter(tree_id=parent.tree_id, lft__range=(parent.lft, parent.rgt - 1))

    def get_descendants(self, include_self=False):
        """
        :returns: A queryset of all the node's descendants as DFS, doesn't
            include the node itself if `include_self` is `False`
        """
        if include_self:
            return self.__class__.get_tree(self)
        if self.is_leaf():
            return self.tree_model().objects.none()
        return self.__class__.get_tree(self).exclude(pk=self.pk)

    def get_descendant_count(self):
        """:returns: the number of descendants of a node."""
        return (self.rgt - self.lft - 1) / 2

    def get_ancestors(self):
        """
        :returns: A queryset containing the current node object's ancestors,
            starting by the root node and descending to the parent.
        """
        if self.is_root():
            return self.tree_model().objects.none()
        return self.tree_model().objects.filter(tree_id=self.tree_id, lft__lt=self.lft, rgt__gt=self.rgt)

    def is_descendant_of(self, node):
        """
        :returns: ``True`` if the node if a descendant of another node given
            as an argument, else, returns ``False``
        """
        return self.tree_id == node.tree_id and self.lft > node.lft and self.rgt < node.rgt

    def get_parent(self, update=False):
        """
        :returns: the parent node of the current node object.
            Caches the result in the object itself to help in loops.
        """
        if self.is_root():
            return
        try:
            if update:
                del self._cached_parent_obj
            else:
                return self._cached_parent_obj
        except AttributeError:
            pass
        # parent = our most direct ancestor
        self._cached_parent_obj = self.get_ancestors().reverse()[0]
        return self._cached_parent_obj

    @classmethod
    def get_root_nodes(cls):
        """:returns: A queryset containing the root nodes in the tree."""
        return cls.tree_model().objects.filter(lft=1)

    class Meta:
        """Abstract model."""

        abstract = True



"""Django admin support for treebeard"""

import sys

from django.contrib import admin, messages
from django.core.exceptions import PermissionDenied
from django.db import transaction
from django.http import HttpResponse, HttpResponseBadRequest
from django.urls import path
from django.utils.encoding import force_str
from django.utils.translation import gettext_lazy as _

from treebeard.al_tree import AL_Node
from treebeard.exceptions import InvalidMoveToDescendant, InvalidPosition, MissingNodeOrderBy, PathOverflow


def check_empty_dict(GET_dict):
    """
    Returns True if the GET query string contains no values, but it can contain
    empty keys.
    This is better than doing not bool(request.GET) as an empty key will return True
    """
    for k, v in GET_dict.items():
        # Don't disable on p(age) or 'all' GET param
        if v and (k not in ["p", "all"]):
            return False
    return True


class TreeAdmin(admin.ModelAdmin):
    """Django Admin class for treebeard."""

    change_list_template = "admin/tree_change_list.html"

    def get_queryset(self, request):
        if issubclass(self.model, AL_Node):
            # AL Trees return a list instead of a QuerySet for .get_tree()
            # So we're returning the regular .get_queryset cause we will use
            # the old admin
            return super().get_queryset(request)

        # We deliberately don't use `get_tree()` here because we want the specific
        # model for inherited models. This assumes that all implementations
        # return the queryset in DFS order (except AL_Node which is handled above).
        return self.model.objects.all()

    def changelist_view(self, request, extra_context=None):
        if issubclass(self.model, AL_Node):
            # For AL trees, use the old admin display
            self.change_list_template = "admin/tree_list.html"

        if extra_context is None:
            extra_context = {}

        extra_context["has_change_permission"] = self.has_change_permission(request)
        extra_context["filtered"] = not check_empty_dict(request.GET)
        return super().changelist_view(request, extra_context)

    def _changeform_view(self, *args, **kwargs):
        # Because Treebeard frequently needs to modify many objects in a tree when one node
        # is added/updated, the normal behaviour of relying on `commit=False` to create
        # unsaved objects before validating inlines etc doesn't work: Treebeard has already
        # made database changes to prepare to insert/move a node.
        # For this reason, if the form has error
        response = super()._changeform_view(*args, **kwargs)
        if getattr(response, "context_data", {}).get("errors", None):
            # There was an error somewhere, likely in an inline, so we'll need to roll back
            transaction.set_rollback(True)
        return response

    def get_urls(self):
        """
        Adds a url to move nodes to this admin
        """
        urls = super().get_urls()
        from django.views.i18n import JavaScriptCatalog

        jsi18n_url = path("jsi18n/", JavaScriptCatalog.as_view(packages=["treebeard"]), name="javascript-catalog")

        new_urls = [
            path(
                "move/",
                self.admin_site.admin_view(self.move_node),
            ),
            jsi18n_url,
        ]
        return new_urls + urls

    def get_node(self, node_id):
        return self.model.objects.get(pk=node_id)

    def try_to_move_node(self, as_child, node, pos, request, target):
        try:
            node.move(target, pos=pos)
            # Call the save method on the (reloaded) node in order to trigger
            # possible signal handlers etc.
            node = self.get_node(node.pk)
            node.save()
        except (MissingNodeOrderBy, PathOverflow, InvalidMoveToDescendant, InvalidPosition):
            e = sys.exc_info()[1]
            # An error was raised while trying to move the node, then set an
            # error message and return 400, this will cause a reload on the
            # client to show the message
            messages.error(request, _("Exception raised while moving node: %s") % _(force_str(e)))
            return HttpResponseBadRequest("Exception raised during move")
        if as_child:
            msg = _('Moved node "%(node)s" as child of "%(other)s"')
        else:
            msg = _('Moved node "%(node)s" as sibling of "%(other)s"')
        messages.info(request, msg % {"node": node, "other": target})
        return HttpResponse("OK")

    def move_node(self, request):
        try:
            node_id = request.POST["node_id"]
            target_id = request.POST["sibling_id"]
            as_child = bool(int(request.POST.get("as_child", 0)))
        except (KeyError, ValueError):
            # Some parameters were missing return a BadRequest
            return HttpResponseBadRequest("Malformed POST params")

        node = self.get_node(node_id)

        if not self.has_change_permission(request, node):
            # The JS will trigger a page reload on error. This message will be displayed after reload.
            messages.error(request, _("You do not have permission to change this object."))
            raise PermissionDenied

        target = self.get_node(target_id)
        is_sorted = True if node.node_order_by else False

        pos = {
            (True, True): "sorted-child",
            (True, False): "last-child",
            (False, True): "sorted-sibling",
            (False, False): "left",
        }[as_child, is_sorted]
        return self.try_to_move_node(as_child, node, pos, request, target)


def admin_factory(form_class):
    """Dynamically build a TreeAdmin subclass for the given form class.

    :param form_class:
    :return: A TreeAdmin subclass.
    """
    return type(form_class.__name__ + "Admin", (TreeAdmin,), dict(form=form_class))



"""Convert strings to numbers and numbers to strings.

Adapted and simplified for Treebeard from https://tabo.pe/projects/numconv/ by Gustavo Picon
"""

# from april fool's rfc 1924
BASE85 = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz!#$%&()*+-;<=>?@^_`{|}~"


class NumConv:
    """Class to create converter objects.

    :param alphabet: A string that will be used as a encoding alphabet.

        The base for conversion is the length of the alphabet.

       The default value is :data:`numconv.BASE85`

    :raise ValueError: when *alphabet* has duplicated characters
    """

    def __init__(self, alphabet: str = BASE85):
        """basic validation and cached_map storage"""
        if len(set(alphabet)) != len(alphabet):
            raise ValueError(f"duplicate characters found in '{alphabet}'")

        self.radix = len(alphabet)
        self.alphabet = alphabet
        self.cached_map = dict(zip(self.alphabet, range(self.radix)))

    def int2str(self, num: int) -> str:
        """Converts an integer into a string.

        :param num: A numeric value to be converted to a string.

        :raise ValueError: when *num* isn't positive
        """
        if num < 0:
            raise ValueError("number must be positive")

        ret = ""
        while True:
            ret = self.alphabet[num % self.radix] + ret
            if num < self.radix:
                return ret
            num //= self.radix

    def str2int(self, chars) -> int:
        """Converts a string into an integer.

        :param chars: A string that will be converted to an integer.

        :raise ValueError: when *chars* is invalid
        """
        ret = 0
        for char in chars:
            try:
                ret = ret * self.radix + self.cached_map[char]
            except KeyError:
                raise ValueError(f"invalid literal for str2int(): '{chars}")
        return ret



import os
import uuid

from django.db import models

from treebeard.al_tree import AL_Node
from treebeard.ltree import LT_Node
from treebeard.mp_tree import MP_Node
from treebeard.ns_tree import NS_Node


class DescMixin(models.Model):
    """
    Model with desc field, handy for identifying objects in tests
    """

    desc = models.CharField(max_length=255)

    def __str__(self):
        return self.desc

    class Meta:
        abstract = True


class RelatedModel(DescMixin): ...


class MP_TestNode(MP_Node, DescMixin):
    steplen = 3


class MP_TestNodeSomeDep(models.Model):
    node = models.ForeignKey(MP_TestNode, on_delete=models.CASCADE)


class MP_TestNodeRelated(MP_Node, DescMixin):
    steplen = 3
    related = models.ForeignKey(RelatedModel, on_delete=models.CASCADE)


class MP_TestNodeInherited(MP_TestNode):
    extra_desc = models.CharField(max_length=255)


class MP_TestNodeCustomId(MP_Node, DescMixin):
    steplen = 3
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)


class NS_TestNode(NS_Node, DescMixin): ...


class NS_TestNodeSomeDep(models.Model):
    node = models.ForeignKey(NS_TestNode, on_delete=models.CASCADE)


class NS_TestNodeRelated(NS_Node, DescMixin):
    related = models.ForeignKey(RelatedModel, on_delete=models.CASCADE)


class NS_TestNodeInherited(NS_TestNode):
    extra_desc = models.CharField(max_length=255)


class AL_TestNode(AL_Node, DescMixin):
    parent = models.ForeignKey(
        "self",
        related_name="children_set",
        null=True,
        db_index=True,
        on_delete=models.CASCADE,
    )
    sib_order = models.PositiveIntegerField()


class AL_TestNodeSomeDep(models.Model):
    node = models.ForeignKey(AL_TestNode, on_delete=models.CASCADE)


class AL_TestNodeRelated(AL_Node, DescMixin):
    parent = models.ForeignKey(
        "self",
        related_name="children_set",
        null=True,
        db_index=True,
        on_delete=models.CASCADE,
    )
    sib_order = models.PositiveIntegerField()
    related = models.ForeignKey(RelatedModel, on_delete=models.CASCADE)


class AL_TestNodeInherited(AL_TestNode):
    extra_desc = models.CharField(max_length=255)


class MP_TestNodeSorted(MP_Node, DescMixin):
    steplen = 1
    node_order_by = ["val1", "val2", "-desc"]
    val1 = models.IntegerField()
    val2 = models.IntegerField()


class MP_TestNodeInheritedSorted(MP_TestNodeSorted): ...


class NS_TestNodeSorted(NS_Node, DescMixin):
    node_order_by = ["val1", "val2", "-desc"]
    val1 = models.IntegerField()
    val2 = models.IntegerField()


class AL_TestNodeSorted(AL_Node, DescMixin):
    parent = models.ForeignKey(
        "self",
        related_name="children_set",
        null=True,
        db_index=True,
        on_delete=models.CASCADE,
    )
    node_order_by = ["val1", "val2", "-desc"]
    val1 = models.IntegerField()
    val2 = models.IntegerField()


class MP_TestNodeAlphabet(MP_Node):
    steplen = 2
    numval = models.IntegerField()


class MP_TestNodeSmallStep(MP_Node):
    steplen = 1
    alphabet = "0123456789"


class MP_TestNodeSortedAutoNow(MP_Node, DescMixin):
    created = models.DateTimeField(auto_now_add=True)
    node_order_by = ["created"]


class MP_TestNodeShortPath(MP_Node, DescMixin):
    steplen = 1
    alphabet = "012345678"


class MP_TestNodeUuid(MP_Node, DescMixin):
    steplen = 1
    custom_id = models.UUIDField(primary_key=True, default=uuid.uuid1, editable=False)


# This is how you change the default fields defined in a Django abstract class
# (in this case, MP_Node), since Django doesn't allow overriding fields, only
# mehods and attributes
MP_TestNodeShortPath._meta.get_field("path").max_length = 4


class MP_TestNode_Proxy(MP_TestNode):
    class Meta:
        proxy = True


class NS_TestNode_Proxy(NS_TestNode):
    class Meta:
        proxy = True


class AL_TestNode_Proxy(AL_TestNode):
    class Meta:
        proxy = True


class MP_TestSortedNodeShortPath(MP_Node, DescMixin):
    steplen = 1
    alphabet = "012345678"
    node_order_by = ["desc"]


MP_TestSortedNodeShortPath._meta.get_field("path").max_length = 4


BASE_MODELS = [
    AL_TestNode,
    MP_TestNode,
    NS_TestNode,
    MP_TestNodeUuid,
    MP_TestNodeCustomId,
]

PROXY_MODELS = [AL_TestNode_Proxy, MP_TestNode_Proxy, NS_TestNode_Proxy]
SORTED_MODELS = [AL_TestNodeSorted, MP_TestNodeSorted, NS_TestNodeSorted]
MP_SHORTPATH_MODELS = [MP_TestNodeShortPath, MP_TestSortedNodeShortPath]
RELATED_MODELS = [AL_TestNodeRelated, MP_TestNodeRelated, NS_TestNodeRelated]

# Pairs of dependent models and base models that they depend on
DEP_MODELS = [(AL_TestNode, AL_TestNodeSomeDep), (MP_TestNode, MP_TestNodeSomeDep), (NS_TestNode, NS_TestNodeSomeDep)]

# Pairs of base models and models that inherit from them
INHERITED_MODELS = [
    (AL_TestNode, AL_TestNodeInherited),
    (MP_TestNode, MP_TestNodeInherited),
    (NS_TestNode, NS_TestNodeInherited),
]

INHERITED_MODELS_WITH_SORT = [
    (MP_TestNodeSorted, MP_TestNodeInheritedSorted),
]

if os.environ.get("DATABASE_ENGINE", "") == "psql":

    class LT_TestNode(LT_Node, DescMixin): ...

    class LT_TestNode_Proxy(LT_TestNode):
        class Meta:
            proxy = True

    class LT_TestNodeSorted(LT_Node, DescMixin):
        node_order_by = ["val1", "val2", "-desc"]
        val1 = models.IntegerField()
        val2 = models.IntegerField()

    class LT_TestNodeInherited(LT_TestNode):
        extra_desc = models.CharField(max_length=255)

    class LT_TestNodeInheritedSorted(LT_TestNodeSorted): ...

    class LT_TestNodeSomeDep(models.Model):
        node = models.ForeignKey(LT_TestNode, on_delete=models.CASCADE)

    BASE_MODELS.append(LT_TestNode)
    PROXY_MODELS.append(LT_TestNode_Proxy)
    DEP_MODELS.append((LT_TestNode, LT_TestNodeSomeDep))
    INHERITED_MODELS.append((LT_TestNode, LT_TestNodeInherited))
    SORTED_MODELS.append(LT_TestNodeSorted)
    INHERITED_MODELS_WITH_SORT.append((LT_TestNodeSorted, LT_TestNodeInheritedSorted))
